// Copyright (C) 2012  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_IOSOCkSTREAM_H_h_
#define DLIB_IOSOCkSTREAM_H_h_

#include "iosockstream/iosockstream.h"


#endif // DLIB_IOSOCkSTREAM_H_h_


